from parser_base import *
from mel_ast import *


class MELError(Exception):
    """Класс для исключений интерпретатора"""

    def __init__(self, message):
        super().__init__(message)


class MELParser(BaseParser):
    """Простейший интерпретатор MEL (Math Expr Lang).
    Вычисления производятся в процессе разбора

    Грамматика:
        num     -> <число>
        ident   -> <идентификатор>
        group   -> num | '(' expr ')'
        mult    -> group (('*' | '/') group)*
        add     -> mult (('+' | '-') mult)*
        expr    -> add
        start   -> expr

    Грамматика 2:
        num     -> <число>
        ident   -> <идентификатор>
        group   -> num | ident | '(' expr ')'
        mult    -> group (('*' | '/') group)*
        add     -> mult (('+' | '-') mult)*
        expr    -> add
        stmt    -> 'read' ident | 'print' expr | ident '=' expr
        prog    -> stmt*
        start   -> prog
    """

    def __init__(self, text: str):
        """Конструктор
        :param text: текст программы
        """
        super().__init__(text)

    def num(self) -> NumNode:
        """num   -> <число>"""
        self.ws()
        pos = self.pos
        while self.curr.isdecimal() or self.curr == '.':
            self.pos += 1
        res = NumNode(float(self.text[pos:self.pos]))
        self.ws()
        return res

    def ident(self) -> IdentNode:
        """ ident   -> <идентификатор> """
        self.ws()
        pos = self.pos
        while self.curr.isalnum():
            self.pos += 1
        res = IdentNode(self.text[pos:self.pos])
        self.ws()
        return res

    def group(self) -> ExprNode:
        """ group   -> num | '(' expr ')' """
        if self.is_parse('('):
            self.parse('(')
            res = self.expr()
            self.parse(')')
            return res
        elif self.curr.isdecimal() or self.curr == '.':
            return self.num()
        else:
            return self.ident()

    def mult(self) -> BinOpNode:
        """ mult    -> group (('*' | '/') group)* """
        res = self.group()
        while self.is_parse('*', '/'):
            op = self.parse('*', '/')
            tmp = self.group()
            res = BinOpNode(BinOp(op), res, tmp)
        return res

    def add(self) -> BinOpNode:
        """ add     -> mult (('+' | '-') mult)* """
        res = self.mult()
        while self.is_parse('+', '-'):
            op = self.parse('+', '-')
            tmp = self.mult()
            res = BinOpNode(BinOp(op), res, tmp)
        return res

    def expr(self) -> ExprNode:
        """ expr    -> add """
        return self.add()

    def stmt(self) -> StmtNode:
        """ stmt    -> 'read' ident | 'print' expr | ident '=' expr """
        if self.is_parse('read', 'print'):
            op = self.parse('read', 'print')
            if op == 'read':
                return ReadNode(self.ident())
            else:
                return PrintNode(self.expr())
        else:
            pos = self.pos
            ident = self.ident()
            self.parse('=')
            return AssignNode(ident, self.expr())

    def prog(self) -> StmtListNode:
        """ prog    -> stmt* """
        stmts = []
        self.ws()
        while self.curr != '$':
            stmts.append(self.stmt())
        return StmtListNode(*stmts)

    def start(self) -> StmtListNode:
        """ start   -> prog """
        return self.prog()

    @classmethod
    def parse_prog(cls, prog: str) -> StmtListNode:
        parser = cls(str(prog))
        return parser.start()
